/* M_CAN test cases - help functions */
 
/* Copyright (c) 2015, Robert Bosch GmbH, Gerlingen, Germany
*  All rights reserved.
*  
*  Redistribution and use in source and binary forms, with or without
*  modification, are permitted provided that the following conditions are met:
*     1. Redistributions of source code must retain the above copyright
*        notice, this list of conditions and the following disclaimer.
*     2. Redistributions in binary form must reproduce the above copyright
*        notice, this list of conditions and the following disclaimer in the
*        documentation and/or other materials provided with the distribution.
*     3. Neither the name of the Robert Bosch GmbH nor the
*        names of its contributors may be used to endorse or promote products
*        derived from this software without specific prior written permission.
*  
*  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
*  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
*  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
*  DISCLAIMED. IN NO EVENT SHALL the Robert Bosch GmbH BE LIABLE FOR ANY
*  DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
*  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
*  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
*  ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
*  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
*  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

/*
   Note: This software is snapshot of an ongoing project and serves as example.
*/

#include "../global_includes.h"

extern can_global_struct global;
#if __XILINX
	extern XIntc intc; // the Xilinx interrupt controller
#endif // __XILINX

// data structure to hold messages after reception via CAN rx_node, RX processing picks up messages from here
rx_msg_list_struct    rx_msg_list[CAN_NUMBER_OF_PRESENT_NODES];
tx_event_list_struct  tx_event_list[CAN_NUMBER_OF_PRESENT_NODES];

/* enable/disable M_CAN nodes in global variable */
void node_set_reset_ena_in_global_struct(const int used_can_nodes[], int used_can_nodes_number)
{
	// Task: set/reset the enable flag in the global structure that represents the state/settings of the M_CAN nodes

	int i, j;

	/* enable/disable the nodes used in this test case */
	for (i = 0; i < CAN_NUMBER_OF_PRESENT_NODES; i++) {
		// disable node by default
		global.can[i].ena = FALSE;

		// enable node, if it is used
	    for (j = 0; j < used_can_nodes_number; j++) {
			// Check if current M_CAN node (i) is used (means, is part of the array used_can_nodes_number)
			if (used_can_nodes[j] == i) {
				// enable this node, as it is used
				global.can[i].ena = TRUE;
			}
	    }
	} // for

} // node_set_reset_ena_in_global_struct()



/* compare messages for equal contents */
boolean msg_compare(can_msg_struct *msg1_ptr, can_msg_struct *msg2_ptr)
{
	int i, payload_bytes;

	/* compare message contents for equality */
	if ( (msg1_ptr->idtype == msg2_ptr->idtype) &&
		 (msg1_ptr->id     == msg2_ptr->id)     &&
	     (msg1_ptr->dlc    == msg2_ptr->dlc)
       )
	{
		// EQUAL Message ID, DLC

	    // compare Messsage Data (payload) for equality
		payload_bytes = convert_DLC_to_data_length(msg1_ptr->dlc);
		for (i = 0; i < payload_bytes; ++i)
		{
			if (msg1_ptr->data[i] != msg2_ptr->data[i])
				return FALSE;
		}

	} else {
		// UN-EQUAL Message ID, DLC
		return FALSE;
	}

    /* compare transmission format for equality */
	if ( (msg1_ptr->fdf == msg2_ptr->fdf) ) {
		// if fdf bit is equal, check if brs is set correctly
		if (msg1_ptr->fdf) {
			if (msg1_ptr->brs != msg2_ptr->brs) {
				// UN-EQUAL BRS
				return FALSE;
			}
		} else {
			// no check - makes no sense
			// because the M_CAN ignores the BRS bit for transmission if FDF==FALSE,
			// consequently checking something that was ignored by the M_CAN and during transmission with the BRS in the received message makes no sense
		}
	} else {
		// UN-EQUAL FDF
		return FALSE;
	}


	/* compare ESI Bit
	 *   comparison makes no sense, since we can control the ESI Bit in a TX Message only in case,
	 *   when the protocol controller is NOT in ERROR_PASSIVE or BUS_OFF state
	 */

	/* at this point in the function: the messages are equal */
	return TRUE;
}


/* Check, if there is any Pending TX Message in the given M_CAN Node */
boolean is_any_tx_msg_request_pending(can_struct *can_ptr)
{
	// Tasks: - Return TRUE:  if there is any pending message  in this M_CAN Node
	//        - Return FALSE: if there are NO pending messages in this M_CAN Node

	M_CAN_TXBRP_union txbrp_reg;
	txbrp_reg.value = reg_get(can_ptr, ADR_M_CAN_TXBRP);

	// check if any TX Buffer Requests Pending
	if ( txbrp_reg.bits.TRP ) {
		// Pending TX Buffer Requests
		return TRUE;
	} else {
		// NO Pending TX Buffer Requests
		return FALSE;
	}

} // is_any_tx_msg_request_pending()


/* Check, if there is any Pending TX Message in any M_CAN Node */
boolean is_any_tx_msg_request_pending_in_any_m_can()
{
	// Tasks: - Return TRUE:  if there is any pending message  in any M_CAN Node
	//        - Return FALSE: if there are NO pending messages in any M_CAN Node

	int can_id;

	// for all M_CAN Nodes
    for (can_id = 0; can_id < CAN_NUMBER_OF_PRESENT_NODES; can_id++) {
    	// check if any TX Buffer Requests Pending in current M_CAN Nodes
    	if ( is_any_tx_msg_request_pending(&global.can[can_id]) ) {
    		// TX Buffer Requests Pending
    		return TRUE;
    	}
    }

    // at this point: NO Pending TX Buffer Requests
	return FALSE;
} // is_any_tx_msg_request_pending_in_any_m_can()


/* Check, if there is any RX Message in the RX FIFO_0 of the given M_CAN Node */
unsigned int is_msg_in_rx_fifo0(can_struct *can_ptr)
{
	// Task: Return the number of messages in the RX FIFO 0
	M_CAN_RXF0S_union rxf0s_reg;

	// read RX FIFO 0 Status
	rxf0s_reg.value = reg_get(can_ptr, ADR_M_CAN_RXF0S);

	// return the FIFO Fill Level
	return rxf0s_reg.bits.F0FL;
} // is_msg_in_rx_fifo0()


/* Check, if there is any RX Message in any M_CAN Node's RX FIFO_0 */
boolean is_msg_in_rx_fifo0_of_any_m_can()
{
	// Task: - Return TRUE:  if there is any RX message  in any M_CAN Node's RX FIFO_0
	//       - Return FALSE: if there is NO

	int can_id;

	// for all M_CAN Nodes
    for (can_id = 0; can_id < CAN_NUMBER_OF_PRESENT_NODES; can_id++) {
    	// check if any TX Buffer Requests Pending in current M_CAN Nodes
    	if ( is_msg_in_rx_fifo0(&global.can[can_id]) ) {
    		// TX Buffer Requests Pending
    		return TRUE;
    	}
    }

    // at this point: NO RX Messages in RX FIFO_0
	return FALSE;
} // is_msg_in_rx_fifo0_of_any_m_can()


/* Message List ################################################# */

// MESSAGE LIST PROPERTIES:
//   list empty:       write_ptr == read_ptr  AND  full= FALSE
//   list full:        write_ptr == read_ptr  AND  full==TRUE
//   list partly full: write_ptr != read_ptr
//   who accesses the list?
//   - ISR:              adds elements    and updates list flags
//   - user code:        removes elements and updates list flags

/* Reset the RX MSG List */
void rx_msg_list_reset()
{
	/* Set internal "pointers" to consistent values - as needed for an empty list */
	int i;

	for (i = 0; i < CAN_NUMBER_OF_PRESENT_NODES; ++i) {
		rx_msg_list[i].next_free_elem = 0;
		rx_msg_list[i].last_full_elem = 0;
		rx_msg_list[i].full = 0;
	}
}


/* Copy Message from Message RAM to Message List */
void copy_msg_from_msg_ram_to_msg_list(can_struct *can_ptr, dword msg_addr_in_msg_ram, rx_info_struct rx_info)
{
	// Tasks: - Alarm: List full
	//        - copy msg to rx_msg_list

	// check if list is full
	if ( msg_list_is_full(can_ptr) )
	{
		if (ERROR_PRINT) printf("[ERROR] Message List of M_CAN_%d is FULL! Message is dropped.\n", can_ptr->id);
		return;
	}

	// list is not full - at this point

	// copy message from CAN node to RX message list element
	m_can_copy_msg_from_msg_ram(can_ptr, msg_addr_in_msg_ram, msg_list_get_next_free_element(can_ptr), rx_info);

	// Add Message to Tail of Message List
	msg_list_add_msg_at_tail(can_ptr);
}


/* Get Head Message from Message List */
can_msg_struct * msg_list_get_head_msg(can_struct *can_ptr)
{
	// Tasks: - Return Head Message from List (as Pointer)
	//        - Does NOT modify list

	// check if list is empty (there must be at least one element!)
	if ( msg_list_is_empty(can_ptr) ) {
		// list empty: FULL STOP;
		if (ERROR_PRINT) printf("[Error] access to emtpy list! (M_CAN_%d)\n", can_ptr->id);
		while (TRUE) {/* do nothing */};
	}

	// list non-empty
	return &rx_msg_list[can_ptr->id].msg[rx_msg_list[can_ptr->id].last_full_elem];
} // get_head_msg_from_msg_list()


/* Remove Head Message from Message List */
void msg_list_remove_head_msg(can_struct *can_ptr)
{
	// Tasks: - Update List Status

	// Hint:  This is the only function removing messages from the list.
	//        This function is only called from normal user code and may be interrupted by an ISR => disable IRQ.

	// check if list is empty (there must be at least one element!)
	if ( msg_list_is_empty(can_ptr) ) {
		// list empty: FULL STOP;
		if (ERROR_PRINT) printf("[Error] access to emtpy list! (M_CAN_%d)\n", can_ptr->id);
		while (TRUE) {/* do nothing */};
	}

	// disable Interrupt Source to prohibit list modification by IR Handler
	#if __ALTERA
		alt_ic_irq_disable(BOSCH_AVALON_M_CAN_IRQ_INTERRUPT_CONTROLLER_ID, BOSCH_AVALON_M_CAN_IRQ);
	#elif __XILINX
		XIntc_Disable(&intc, XPAR_XPS_INTC_0_M_CAN_PLB_DE_0_M_CAN_INT_INTR);
	#endif // __XILINX

	// progress in list - keep order of steps  to avoid inconsistency ========
	// This code is not Interrupted!

	// 1. OPERATION
	// increment last_full_elem (remove head msg)
	if ( rx_msg_list[can_ptr->id].last_full_elem == (RX_MESSAGE_LIST_ENTRIES-1) ) {
		rx_msg_list[can_ptr->id].last_full_elem = 0;
	} else {
		rx_msg_list[can_ptr->id].last_full_elem++;
	}

	// 2. OPERATION - Update full flag
	// If full flag is set, reset it, since an elemente was removed from list
	if (rx_msg_list[can_ptr->id].full) {
		rx_msg_list[can_ptr->id].full = FALSE;
	}

	// enable Interrupt Source (that was disable before)
	#if __ALTERA
		alt_ic_irq_enable(BOSCH_AVALON_M_CAN_IRQ_INTERRUPT_CONTROLLER_ID, BOSCH_AVALON_M_CAN_IRQ);
	#elif __XILINX
		XIntc_Enable(&intc, XPAR_XPS_INTC_0_M_CAN_PLB_DE_0_M_CAN_INT_INTR);
	#endif // __XILINX

} // remove_head_msg_from_msg_list()


/* Get Next Free Element Pointer of Message List */
can_msg_struct * msg_list_get_next_free_element(can_struct *can_ptr)
{
	// Tasks: - Return the pointer to the next free element
	//        - Does NOT modify list

	// check if list is full (there must be at least one free element!)
	if ( msg_list_is_full(can_ptr) ) {
		// list full: FULL STOP;
		if (ERROR_PRINT) printf("[Error] get_next_free_element_from_msg_list(), but list is full! Full Stop here! (M_CAN_%d)\n", can_ptr->id);
		while (TRUE) {/* do nothing */};
	}

	// list not full
	return &rx_msg_list[can_ptr->id].msg[rx_msg_list[can_ptr->id].next_free_elem];
} // get_next_free_element_from_msg_list()


/* Add Message to Tail of Message List */
void msg_list_add_msg_at_tail(can_struct *can_ptr)
{
	// Tasks: - Update List Status, i.e. progress with next_free_element pointer
	//        - The current next_free_elem has to already contain the msg added,
	//          do msg copy before calling this function

	// Hint:  This is the only function adding messages to the list.
	//        This function is only called from an ISR, i.e. it is not interrupted.

	// check if list is full (there must be at least one free element!)
	if ( msg_list_is_full(can_ptr) ) {
		// list full: FULL STOP;
		if (ERROR_PRINT) printf("[Error] get_next_free_element_from_msg_list(), but list is full! Full Stop here! (M_CAN_%d)\n", can_ptr->id);
		while (TRUE) {/* do nothing */};
	}

	// increment list pointer to next free element in List
	if ( rx_msg_list[can_ptr->id].next_free_elem == (RX_MESSAGE_LIST_ENTRIES-1) ) {
		rx_msg_list[can_ptr->id].next_free_elem = 0;
	} else {
		rx_msg_list[can_ptr->id].next_free_elem++;
	}

	// check if list got now full
	if ( rx_msg_list[can_ptr->id].next_free_elem == rx_msg_list[can_ptr->id].last_full_elem) {
		rx_msg_list[can_ptr->id].full = TRUE;
	}
} // add_msg_to_tail_of_msg_list()


/* Check, if Message List is EMPTY */
boolean msg_list_is_empty(can_struct *can_ptr)
{
	// Tasks: - Return List Status: TRUE=> EMPTY, FALSE=> non-EMPTY

	if ( (rx_msg_list[can_ptr->id].last_full_elem == rx_msg_list[can_ptr->id].next_free_elem) &&
		 (rx_msg_list[can_ptr->id].full == FALSE)
	   ) {
		// list empty
		return TRUE;
	} else {
		// list non-empty
		return FALSE;
	}

} // is_empty_msg_list()


/* Check, if Message List is FULL */
boolean msg_list_is_full(can_struct *can_ptr)
{
	// Tasks: - Return List Status: TRUE=> FULL, FALSE=> non-FULL
	return rx_msg_list[can_ptr->id].full;
} // is_full_msg_list()


/* Check, if there is any RX Message List in any of the Message Lists */
boolean is_any_msg_in_any_rx_msg_list()
{
	// Tasks: - Return all Lists Status:
	//          TRUE  => any Msg available,
	//          FALSE =>  no Msg available

	int can_id;

	// for all Lists (each M_CAN has one list)
    for (can_id = 0; can_id < CAN_NUMBER_OF_PRESENT_NODES; can_id++) {
    	// check if list is non-empty
    	if (msg_list_is_empty(&global.can[can_id]) == FALSE) {
    		// list non-empty
    		return TRUE;
    	}
    }

    // at this point: lists are empty
	return FALSE;

} // is_empty_msg_list()



/* Tx Event List ################################################# */

// TX EVENT FIFO LIST PROPERTIES:
//   list empty:       write_ptr == read_ptr  AND  full= FALSE
//   list full:        write_ptr == read_ptr  AND  full==TRUE
//   list partly full: write_ptr != read_ptr
//   who accesses the list?
//   - ISR:              adds elements    and updates list flags
//   - user code:        removes elements and updates list flags

/* Reset the Tx Event FIFO list */
void tx_event_list_reset()
{
	/* Description: Set internal "pointers" to consistent values - as needed for an empty list */
	int i;

	for (i = 0; i < CAN_NUMBER_OF_PRESENT_NODES; ++i) {
		tx_event_list[i].next_free_elem = 0;
		tx_event_list[i].last_full_elem = 0;
		tx_event_list[i].full = 0;
	}
}


/* Copy a Tx Event FIFO element from Message RAM to TX Event List */
void copy_event_element_from_msg_ram_to_tx_event_list(can_struct *can_ptr, dword tx_event_element_addr_in_msg_ram)
{
	/* Description: copies a Tx Event FIFO element from Message RAM and adds it to tx_event_list
	 *
	 * Parameters
	 * 	- can_ptr: pointer to the M_CAN node, where the configuration should be applied
	 * 	- msg_addr_in_msg_ram: address of the Tx Event FIFO element in the message RAM
	 */

	tx_event_element_struct * tx_event_fifo_element_ptr;

	// check if list is full
	if ( tx_event_list_is_full(can_ptr) )
	{
		if (ERROR_PRINT) printf("[ERROR] Tx Event Message List of M_CAN_%d is FULL! Message is dropped.\n", can_ptr->id);
		return;
	}
	// list is not full - at this point

	// get the position in the list where the new element can be added
	tx_event_fifo_element_ptr = tx_event_list_get_next_free_element(can_ptr);

	// copy a Tx Event FIFO element from CAN node to tx_event_list
	m_can_copy_tx_event_element_from_msg_ram(can_ptr, tx_event_element_addr_in_msg_ram, tx_event_fifo_element_ptr);

	// Now the tx_event_element is already copied to the free element of the tx_event_list,
	// but the tx_event_list doesn't know that yet (= next_free/last_full status is not yet updated)

	// Add Tx Event Element to Tail of tx_event_list
	tx_event_list_add_element_at_tail(can_ptr); // this only updates the list status
}


/* Get Head element from Tx Event List */
tx_event_element_struct * tx_event_list_get_head_element(can_struct *can_ptr)
{
	/* Description:
	 * 	- Returns Head element from List (as Pointer)
     * 	- Does NOT modify list
     *
	 * Parameters
	 * 	- can_ptr: pointer to the M_CAN node, whose list is being accessed
	 */

	// check if list is empty (there must be at least one element!)
	if ( tx_event_list_is_empty(can_ptr) ) {
		printf("[Error] access to emtpy list! (M_CAN_%d)\n", can_ptr->id);
		// list empty: FULL STOP;
		if (ERROR_PRINT) printf("[Error] access to emtpy list! (M_CAN_%d)\n", can_ptr->id);
		while (TRUE) {/* do nothing */};
	}

	// list non-empty
	return &tx_event_list[can_ptr->id].tx_event_elem[tx_event_list[can_ptr->id].last_full_elem];
}


/* Remove Head element from Tx Event List */
void tx_event_list_remove_head_element(can_struct *can_ptr)
{
	/* Description: Update List Status
     *
	 * Parameters
	 * 	- can_ptr: pointer to the M_CAN node, whose list is being accessed
	 *
	 * Note:  This is the only function removing elements from the list.
	 *	 	  This function is only called from normal user code and may be interrupted by an ISR => disable IRQ.
	 */

	// check if list is empty (there must be at least one element!)
	assert_print_error(tx_event_list_is_empty(can_ptr) != TRUE, "[Error] access to emtpy Tx Event list!");
	//if ( tx_event_list_is_empty(can_ptr) ) {
	//	// list empty: FULL STOP;
	//	if (ERROR_PRINT) printf("[Error] access to emtpy list! (M_CAN_%d)\n", can_ptr->id);
	//	while (TRUE) {/* do nothing */};
	//}

	// disable Interrupt that modifies list
	#if __ALTERA
		alt_ic_irq_disable(BOSCH_AVALON_M_CAN_IRQ_INTERRUPT_CONTROLLER_ID, BOSCH_AVALON_M_CAN_IRQ);
	#elif __XILINX
		XIntc_Disable(&intc, XPAR_XPS_INTC_0_M_CAN_PLB_DE_0_M_CAN_INT_INTR);
	#endif // __XILINX

	// progress in list - keep order of steps  to avoid inconsistency ========
	// This code is not Interrupted!

	// 1. OPERATION
	// increment last_full_elem (remove head msg)
	if ( tx_event_list[can_ptr->id].last_full_elem == (TX_EVENT_FIFO_LIST_ENTRIES-1) ) {
		tx_event_list[can_ptr->id].last_full_elem = 0;
	} else {
		tx_event_list[can_ptr->id].last_full_elem++;
	}


	// 2. OPERATION - Update full flag
	// If full flag is set, reset it, since an elemente was removed from list
	if (tx_event_list[can_ptr->id].full) {
		tx_event_list[can_ptr->id].full = FALSE;
	}

	// enable Interrupt that modifies list
	#if __ALTERA
		alt_ic_irq_enable(BOSCH_AVALON_M_CAN_IRQ_INTERRUPT_CONTROLLER_ID, BOSCH_AVALON_M_CAN_IRQ);
	#elif __XILINX
		XIntc_Enable(&intc, XPAR_XPS_INTC_0_M_CAN_PLB_DE_0_M_CAN_INT_INTR);
	#endif // __XILINX

}


/* Get Next Free Element Pointer of Tx Event List */
tx_event_element_struct * tx_event_list_get_next_free_element(can_struct *can_ptr)
{
	/* Description:
	 * 	- Return the pointer to the next free element of the tx_event_list
	 * 	- Does NOT modify list
	 *
	 * Parameters
	 * 	- can_ptr: pointer to the M_CAN node, whose list is being accessed
	 */

	// check if list is full (there must be at least one free element!)
	if ( tx_event_list_is_full(can_ptr) ) {
		// list full: FULL STOP;
		if (ERROR_PRINT) printf("[Error] tx_event_list_get_next_free_element(), but list is full! Full Stop here! (M_CAN_%d)\n", can_ptr->id);
		while (TRUE) {/* do nothing */};
	}

	// list not full
	return &tx_event_list[can_ptr->id].tx_event_elem[tx_event_list[can_ptr->id].next_free_elem];

}


/* Add Element to Tail of Tx Event List */
void tx_event_list_add_element_at_tail(can_struct *can_ptr)
{
	/* Description:
	 * 	- Update List Status, i.e. progress with next_free_element pointer
	 * 	- The current next_free_elem has to already contain the tx_event_fifo_element added, do element copy before calling this function
	 *
	 * Parameters
	 * 	- can_ptr: pointer to the M_CAN node, whose list is being accessed
	 *
	 * Note:  This is the only function adding element to the list.
	 *        This function is only called from an ISR, i.e. it is not interrupted.
	 */

	// check if list is full (there must be at least one free element!)
	if ( tx_event_list_is_full(can_ptr) ) {
		// list full: FULL STOP;
		if (ERROR_PRINT) printf("[Error] add_msg_to_tail_of_tx_event_list(), but list is full! Full Stop here! (M_CAN_%d)\n", can_ptr->id);
		while (TRUE) {/* do nothing */};
	}

	// increment list pointer to next free element in List
	if ( tx_event_list[can_ptr->id].next_free_elem == (TX_EVENT_FIFO_LIST_ENTRIES-1) ) {
		tx_event_list[can_ptr->id].next_free_elem = 0;
	} else {
		tx_event_list[can_ptr->id].next_free_elem++;
	}

	// check if list got now full
	if ( tx_event_list[can_ptr->id].next_free_elem == tx_event_list[can_ptr->id].last_full_elem) {
		tx_event_list[can_ptr->id].full = TRUE;
	}
}


/* Check, if Tx Event List is EMPTY */
boolean tx_event_list_is_empty(can_struct *can_ptr)
{
	/* Description: Return List Status: TRUE=> EMPTY, FALSE=> non-EMPTY
	 */

	if ( (tx_event_list[can_ptr->id].last_full_elem == tx_event_list[can_ptr->id].next_free_elem) &&
		 (tx_event_list[can_ptr->id].full == FALSE)
	   ) {
		// list empty
		return TRUE;
	} else {
		// list non-empty
		return FALSE;
	}
}


/* Check, if Tx Event List is FULL */
boolean tx_event_list_is_full(can_struct *can_ptr)
{
	// Description:  Return List Status: TRUE=> FULL, FALSE=> non-FULL
	return tx_event_list[can_ptr->id].full;
}


/* Message Generation ################################### */

/* Generate a TX Message - used for simple CAN FD test */
can_msg_struct * m_can_test_simple_generate_tx_message()
{
	static can_msg_struct tx_msg; // variable remains even after this function ended

    const int    min_std_msg_id = 0x0;
    const int    max_std_msg_id = 0x7FF;
    const dword step_std_msg_id = 1<<3;

    const int    min_ext_msg_id = 0x0;
    const int    max_ext_msg_id = 0x1FFFFFFF;
    const dword step_ext_msg_id = 1<<20;

    const byte min_msg_dlc = 1;
    const byte max_msg_dlc = 15;

	static can_id_type_enum current_msg_id_type = standard_id;
	static dword id  = 0x0;
	static byte  dlc = 15;
	static byte  databyte = 0x0;

    int i;

    // message direction: transmit
	tx_msg.direction = tx_dir;

	// tx event fifo settings => do not request event element generation
	tx_msg.efc = FALSE;

    // statically set FD bits
    tx_msg.fdf = TRUE;
    tx_msg.brs = TRUE;
    tx_msg.esi = FALSE;

    // assign current id
	tx_msg.id = id;
	tx_msg.idtype = current_msg_id_type;

    // determine NEXT id
    if (current_msg_id_type == standard_id) {
    	// msg id
        if ((id+step_std_msg_id) < max_std_msg_id) {
        	id = id + step_std_msg_id;
        } else {
        	// switch to extended_id
        	current_msg_id_type = extended_id;
        	id = min_ext_msg_id;
        }
    } else { // extended_id
    	// msg id
        if ((id+step_ext_msg_id) < max_ext_msg_id) {
        	id = id + step_ext_msg_id;
        } else {
        	// switch to standard_id
        	current_msg_id_type = standard_id;
        	id = min_std_msg_id;
        }
    }

    // dlc
    if (dlc < max_msg_dlc) {
    	dlc++;
    } else {
    	dlc = min_msg_dlc;
    }
	tx_msg.dlc = dlc;

    // data field
	databyte++;
    for (i = 0; i < convert_DLC_to_data_length(tx_msg.dlc); ++i) {
		tx_msg.data[i] = databyte;
	}

    return &tx_msg;
}


/* Statistics ########################################### */

/* Update Message Statistics */
void update_msg_statistics_at_mram_access(can_struct *can_ptr, can_msg_struct *msg_ptr)
{
	// Function updates the statistics messages sent (TX) and received (RX)

	// it this is an RX Message
	if (msg_ptr->direction == rx_dir) {

		can_ptr->stat.rx.msgs_mram++;  // increment message counter

	    if (msg_ptr->fdf) {
	    	can_ptr->stat.rx.fdf++;
	    }

	    if (msg_ptr->brs) {
	    	can_ptr->stat.rx.brs++;
	    }

	    if (msg_ptr->esi) {
	    	can_ptr->stat.rx.esi++;
	    }

	} // if RX-Message

	// it this is an TX Message
	if (msg_ptr->direction == tx_dir) {

		can_ptr->stat.tx.msgs_mram++;  // increment message counter

	    if (msg_ptr->fdf) {
	    	can_ptr->stat.tx.fdf++;
	    }

	    if (msg_ptr->brs) {
	    	can_ptr->stat.tx.brs++;
	    }

	    if (msg_ptr->esi) {
	    	can_ptr->stat.tx.esi++;
	    }

	} // if RX-Message


} // update statistics


/* Print Statistics */
void print_statistics(can_struct *can_ptr)
{

	// Hint: As xil_printf does not support the format specifier %u, instad %d is used.

	word tld_mean; // help variable

	printf("Statistics M_CAN_%d\n", can_ptr->id);
	printf("  TX         msgs_mram=%7d,    msgs_ir_ok=%7d,  flags: fdf=%7d,  brs=%7d,  esi=%7d,  tx_evnt_elems=%7d\n\n",
			  can_ptr->stat.tx.msgs_mram,
			  can_ptr->stat.tx.msgs_ir_ok,
			  can_ptr->stat.tx.fdf,
			  can_ptr->stat.tx.brs,
			  can_ptr->stat.tx.esi,
			  can_ptr->stat.tx_event_elements
			  );
	printf("  RX         msgs_mram=%7d,    msgs_ir_ok=%7d,  flags: fdf=%7d,  brs=%7d,  esi=%7d\n",
			  can_ptr->stat.rx.msgs_mram,
			  can_ptr->stat.rx.msgs_ir_ok,
			  can_ptr->stat.rx.fdf,
			  can_ptr->stat.rx.brs,
			  can_ptr->stat.rx.esi
			  );
	printf("  RX  CHECK    msgs_ok=%7d,      msg_dupl=%7d,    msgs_err=%7d\n",
			  can_ptr->stat.rx_check.msgs_ok,
			  can_ptr->stat.rx_check.msgs_duplicate,
			  can_ptr->stat.rx_check.msgs_err
			  );

	// calculate tld_mean, avoid division by zero
	if (can_ptr->stat.tld.sum_cnt == 0) {
		tld_mean = 0;
	} else {
		tld_mean = can_ptr->stat.tld.sum/can_ptr->stat.tld.sum_cnt;
	}

	printf("  TLD              min=%7d,           max=%7d,        mean=%7d,       count=%7d   (TLD values in clk periods)\n",
			  can_ptr->stat.tld.min,
			  can_ptr->stat.tld.max,
			  tld_mean,
			  can_ptr->stat.tld.sum_cnt
			  );
	printf("  Protocol        boff=%7d,         ewarn=%7d,       epass=%7d,         pxe=%7d\n",
			can_ptr->stat.protocol.boff,
			can_ptr->stat.protocol.ewarn,
			can_ptr->stat.protocol.epass,
			can_ptr->stat.protocol.pxe

			);
	printf("           LEC:  Stuff=%7d,          Form=%7d,         Ack=%7d,        Bit1=%7d, Bit0=%7d,  Crc=%7d\n",
			can_ptr->stat.protocol.lec[1],
			can_ptr->stat.protocol.lec[2],
			can_ptr->stat.protocol.lec[3],
			can_ptr->stat.protocol.lec[4],
			can_ptr->stat.protocol.lec[5],
			can_ptr->stat.protocol.lec[6]
			);
	printf("           DLEC: Stuff=%7d,          Form=%7d,         Ack=%7d,        Bit1=%7d, Bit0=%7d,  Crc=%7d\n",
			can_ptr->stat.protocol.dlec[1],
			can_ptr->stat.protocol.dlec[2],
			can_ptr->stat.protocol.dlec[3],
			can_ptr->stat.protocol.dlec[4],
			can_ptr->stat.protocol.dlec[5],
			can_ptr->stat.protocol.dlec[6]
			);
	printf("  HW Access Error  ARA=%7d,           BEU=%7d,         BEC=%7d,        MRAF=%7d\n",
			  can_ptr->stat.hw_access.ara,
			  can_ptr->stat.hw_access.beu,
			  can_ptr->stat.hw_access.bec,
			  can_ptr->stat.hw_access.mraf
			  );

} // function print stat



/* Reset Statistics */
void reset_statistic()
{
	int i;

	for (i = 0; i < CAN_NUMBER_OF_PRESENT_NODES; ++i) {
		// init struct with zeros
		memset(&global.can[i].stat, 0, sizeof(can_statistics_struct));
		global.can[i].stat.tld.min = 9999;  // init with a large value to be able to simply set the first tld_min
	}
}



/* Check the consistency of the Statistics */
void check_statistic_consistency_single_board_test(const int used_can_nodes[], int used_can_nodes_number)
{
	/* Works only with - a single board stress test
	 *                 - all frames are stored by each M_CAN (no message filtering is used)
	 */

	// Hint: As xil_printf does not support the format specifier %u, instad %d is used.

	int i, j, can_id_rx, can_id_tx;
    dword total_tx_msgs_of_all_other_nodes;
    dword fdf_tx_msgs_of_all_other_nodes;
    dword brs_tx_msgs_of_all_other_nodes;
    dword esi_tx_msgs_of_all_other_nodes;

    printf("Statistics Consistency Check\n");

    /* Check Total RX/TX Frame counts */
	for (i = 0; i < used_can_nodes_number; ++i) {
		// select RX Node
		can_id_rx = used_can_nodes[i];

		// sum up all tx messages of the other nodes
		total_tx_msgs_of_all_other_nodes = 0;
		fdf_tx_msgs_of_all_other_nodes = 0;
		brs_tx_msgs_of_all_other_nodes = 0;
		esi_tx_msgs_of_all_other_nodes = 0;

		for (j = 0; j < used_can_nodes_number; ++j) {
			// select M_CAN
			can_id_tx = used_can_nodes[j];
			// don't add the messages of the RX Node
			if (can_id_rx != can_id_tx) {
				total_tx_msgs_of_all_other_nodes += global.can[can_id_tx].stat.tx.msgs_mram;
				fdf_tx_msgs_of_all_other_nodes   += global.can[can_id_tx].stat.tx.fdf;
				brs_tx_msgs_of_all_other_nodes   += global.can[can_id_tx].stat.tx.brs;
				esi_tx_msgs_of_all_other_nodes   += global.can[can_id_tx].stat.tx.esi;
			}
		}

		/* Report on Console */
		// check if received msgs of RX Node == transmitted messages in all other nodes
		printf("  M_CAN_%d", can_id_rx);

		// total
		printf("  Total: RX=%7d, others TX=%7d", global.can[can_id_rx].stat.rx.msgs_mram, total_tx_msgs_of_all_other_nodes);
		if (global.can[can_id_rx].stat.rx.msgs_mram == total_tx_msgs_of_all_other_nodes) {
			printf("        ");
		} else {
			printf(" => MISS");
		}
		// fdf
		printf("  FDF: RX=%7d, others TX=%7d", global.can[can_id_rx].stat.rx.fdf, fdf_tx_msgs_of_all_other_nodes);
		if (global.can[can_id_rx].stat.rx.fdf == fdf_tx_msgs_of_all_other_nodes) {
			printf("        \n");
		} else {
			printf(" => MISS\n");
		}
		// brs
		printf("             BRS: RX=%7d, others TX=%7d", global.can[can_id_rx].stat.rx.brs, brs_tx_msgs_of_all_other_nodes);
		if (global.can[can_id_rx].stat.rx.brs == brs_tx_msgs_of_all_other_nodes) {
			printf("        ");
		} else {
			printf(" => MISS");
		}
		// esi
		printf("  ESI: RX=%7d, others TX=%7d", global.can[can_id_rx].stat.rx.esi, esi_tx_msgs_of_all_other_nodes);
		if (global.can[can_id_rx].stat.rx.esi == esi_tx_msgs_of_all_other_nodes) {
			printf("        \n");
		} else {
			printf(" => MISS\n");
		}
	} // for

} // check_statistic_consistency_single_board_fd_stress_test()


/* Check the statistics for recorded error events */
void check_statistic_for_detected_errors(const int used_can_nodes[], int used_can_nodes_number)
{
	/* Target: - checks all statistics that recored error events
	 *         - if any error event is observed (statistic /= 0), an information is printed out
	 */

	// Hint: As xil_printf does not support the format specifier %u, instad %d is used.

	int i, j, can_id;
    dword total_error_events_of_one_node;

    printf("Statistics Check for recorded Error Events\n");

    /* Check Total RX/TX Frame counts */
	for (i = 0; i < used_can_nodes_number; ++i) {
		// select RX Node
		can_id = used_can_nodes[i];

		// SUM UP ALL ERROR events recorded in statistics counters of ONE node
		// 1: sum up all error protocoll errors
		total_error_events_of_one_node  = global.can[can_id].stat.protocol.boff;
		total_error_events_of_one_node += global.can[can_id].stat.protocol.epass;
		total_error_events_of_one_node += global.can[can_id].stat.protocol.ewarn;
		total_error_events_of_one_node += global.can[can_id].stat.protocol.pxe;

		// add all LEC Errors
		for (j = 1; j < 7; j++) {  // LEC Error are stored in lec[1]..lec[6]
			total_error_events_of_one_node += global.can[can_id].stat.protocol.lec[j];
			total_error_events_of_one_node += global.can[can_id].stat.protocol.dlec[j];
		}

		// add all hardware access errors
		total_error_events_of_one_node  += global.can[can_id].stat.hw_access.ara;
		total_error_events_of_one_node  += global.can[can_id].stat.hw_access.bec;
		total_error_events_of_one_node  += global.can[can_id].stat.hw_access.beu;
		total_error_events_of_one_node  += global.can[can_id].stat.hw_access.mraf;

		// add all RX message check errors
		total_error_events_of_one_node  += global.can[can_id].stat.rx_check.msgs_err;
		total_error_events_of_one_node  += global.can[can_id].stat.rx_check.msgs_duplicate;


		/* Report on Console */
		printf("  M_CAN_%d  Total Errors Recorded: %7d", can_id, total_error_events_of_one_node);
		if (total_error_events_of_one_node != 0) {
			printf(" => ERRORS OCCURRED!");
		}
		printf("\n");

	} // for

} // check_statistic_for_detected_errors()


/* Misc ################################################# */

/* Print Message */
void print_msg(can_msg_struct *msg_ptr)
{
	int i;

	if (msg_ptr->direction == tx_dir) {
		printf(" TX Msg:");
	} else if (msg_ptr->direction == rx_dir) {
		printf(" RX Msg:");
		switch(msg_ptr->rx_info.rx_via)
		{
			case FIFO_0: printf(" via FIFO 0   Index= %2d", msg_ptr->rx_info.buffer_index);
						 break;
			case FIFO_1: printf(" via FIFO 1   Index= %2d", msg_ptr->rx_info.buffer_index);
						 break;
			case DEDICATED_RX_BUFFER : printf(" via Ded buff Index= %2d", msg_ptr->rx_info.buffer_index);
									   break;
		}//switch
	} else {
		printf(" Msg (direction not set!):");
	}

	printf("  IDTYPE= ");
	if (msg_ptr->idtype == standard_id) {
		printf("STD  ID= 0x%03x     ", msg_ptr->id);
	} else {
		printf("EXT  ID= 0x%08x", msg_ptr->id);
	}

	printf("  REMOTE= %d  FDF= %d  BRS= %d  ESI= %d  DLC= %2d  DataBytes[0..%2d]= 0x",
			msg_ptr->remote,
			msg_ptr->fdf,
			msg_ptr->brs,
			msg_ptr->esi,
			msg_ptr->dlc,
			convert_DLC_to_data_length(msg_ptr->dlc));

	for (i = 0; i < convert_DLC_to_data_length(msg_ptr->dlc); ++i) {
		printf(" %02x", msg_ptr->data[i]);
	}

	printf("\n");

} // function print_msg()


/* prints the content of one Tx Event Element */
void print_tx_event_fifo_element(tx_event_element_struct *tx_event_element_ptr)
{
	/* Description: prints the contents of one Tx Event FIFO Element
	 *
	 * Parameters
	 * 	- tx_event_element_ptr: pointer to the Tx Event FIFO element whose contents should be printed
	 */
	printf("  Message Marker= 0x%02x  Time Stamp= 0x%04x",
			tx_event_element_ptr->mm,
			tx_event_element_ptr->txts);

	printf("  IDTYPE= ");
	if (tx_event_element_ptr->idtype == standard_id) {
		printf("STD");
	} else {
		printf("EXT");
	}

	printf("  ID= 0x%08x  ET= 0x%1x  REMOTE= %d  FDF= %d  BRS= %d  ESI= %d  DLC= %2d ",
			tx_event_element_ptr->id,
			tx_event_element_ptr->et,
			tx_event_element_ptr->remote,
			tx_event_element_ptr->fdf,
			tx_event_element_ptr->brs,
			tx_event_element_ptr->esi,
			tx_event_element_ptr->dlc);

	printf("\n");
}


/* Print relative Message RAM Start-Addresses */
void print_mram_startaddresses()
{

	int i;
    dword rel_start_addr_word;

	for (i = 0; i < CAN_NUMBER_OF_PRESENT_NODES; ++i) {
		printf("M_CAN_%d: Message RAM Start Addresses (byte) from M_CAN point of view\n",i);

		// calculate start address, that is relative to Message_RAM Base + convert to word address
		rel_start_addr_word = (global.can[i].mram_sa.SIDFC_FLSSA - MESSAGE_RAM_BASE);
		printf("  SIDFC_FLSSA = 0x%4.4x\n", rel_start_addr_word);
		rel_start_addr_word = (global.can[i].mram_sa.XIDFC_FLESA - MESSAGE_RAM_BASE);
		printf("  XIDFC_FLESA = 0x%4.4x\n", rel_start_addr_word);

		rel_start_addr_word = (global.can[i].mram_sa.RXF0C_F0SA - MESSAGE_RAM_BASE);
		printf("  RXF0C_F0SA  = 0x%4.4x\n", rel_start_addr_word);
		rel_start_addr_word = (global.can[i].mram_sa.RXF1C_F1SA - MESSAGE_RAM_BASE);
		printf("  RXF1C_F1SA  = 0x%4.4x\n", rel_start_addr_word);
		rel_start_addr_word = (global.can[i].mram_sa.RXBC_RBSA - MESSAGE_RAM_BASE);
		printf("  RXBC_RBSA   = 0x%4.4x\n", rel_start_addr_word);

		rel_start_addr_word = (global.can[i].mram_sa.TXEFC_EFSA - MESSAGE_RAM_BASE);
		printf("  TXEFC_EFSA  = 0x%4.4x\n", rel_start_addr_word);
		rel_start_addr_word = (global.can[i].mram_sa.TXBC_TBSA - MESSAGE_RAM_BASE);
		printf("  TXBC_TBSA   = 0x%4.4x\n", rel_start_addr_word);
	}

} // function print_mram_startaddresses()

