/* M_CAN IRQ handling */
 
/* Copyright (c) 2015, Robert Bosch GmbH, Gerlingen, Germany
*  All rights reserved.
*  
*  Redistribution and use in source and binary forms, with or without
*  modification, are permitted provided that the following conditions are met:
*     1. Redistributions of source code must retain the above copyright
*        notice, this list of conditions and the following disclaimer.
*     2. Redistributions in binary form must reproduce the above copyright
*        notice, this list of conditions and the following disclaimer in the
*        documentation and/or other materials provided with the distribution.
*     3. Neither the name of the Robert Bosch GmbH nor the
*        names of its contributors may be used to endorse or promote products
*        derived from this software without specific prior written permission.
*  
*  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
*  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
*  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
*  DISCLAIMED. IN NO EVENT SHALL the Robert Bosch GmbH BE LIABLE FOR ANY
*  DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
*  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
*  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
*  ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
*  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
*  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

/*
   Note: This software is snapshot of an ongoing project and serves as example.
*/

#include "../global_includes.h"

extern can_global_struct global;
//extern rx_msg_list_struct  rx_msg_list[CAN_NUMBER_OF_PRESENT_NODES];



/* Find M_CAN nodes with active IR */
void find_m_can_with_active_IR()
{
	int i;
	dword ir_reg;

	// scan all M_CAN nodes that have an active IRs
    for (i = 0; i < CAN_NUMBER_OF_PRESENT_NODES; ++i) {

    	// M_CAN and M_TTCAN nodes have both the identical IR Register
    	ir_reg = reg_get(&global.can[i], ADR_M_CAN_IR);
    	if (ir_reg != 0) {
    		//if (DEBUG_PRINT) printf("[M_CAN_%d] Interrupt occurred,         IR_reg = 0x%8.8x\n", i, ir_reg);
    		//printf("[M_CAN_%d] Interrupt occurred,         IR_reg = 0x%8.8x\n", i, ir_reg);

    		m_can_process_IRQ(&global.can[i], ir_reg);
    	}

    	// M_TTCAN Nodes have additionally the TTIR Register
    	if (global.can[i].is_m_ttcan_node) {
    		ir_reg = reg_get(&global.can[i], ADR_M_CAN_TTIR);
        	if (ir_reg != 0) {
        		m_can_process_TT_IRQ(&global.can[i], ir_reg);
        	}
    	}

    } // for
}


/* Handle IR of M_CAN Nodes */
void m_can_process_IRQ(can_struct *can_ptr, dword ir_value)
{
	static M_CAN_IR_union ir_reg;
	static M_CAN_IR_union ir_reset_reg;
	static M_CAN_IR_union ir_reset_reg_immediately; // used to reset an interrupt immediately
	M_CAN_PSR_union psr_reg;
	dword tld_last; // last measured transmitter loop delay

	/* Process only Enabled Interrupts => Mask out Interrupt Flags that are not enabled
	 *  - The M_CAN sets the interrupt flags always when the flag condition occurs.
	 *  - Disabling an interrupt via IE (Interrupt Enable) register has no effect on the setting of the interrupt flag in IR register
	 *  - The IE register only decides which interrupt flags in the IR(Interrupt Register) will be signalled on an interrupt line.
	 *  - HERE In this ISR only those interrupts are processed, which are enabled
	 *      - the not enabled interrupts are masked out with help of the interrupt enabled register
	 *      - the interrupt_init function stores in software the inerrupt_enable register written to the M_CAN
	 *      - Masking out makes sense to avoid unintended software execution: i.e. RF0N-flag would be evaluated each time this ISR is called, even if this Interrupt is not enabled
	 */
	ir_reg.value = ir_value & can_ptr->interrupts_enabled;
	
    // read PSR Register once at the beginning of this ISR
	psr_reg.value = reg_get(can_ptr, ADR_M_CAN_PSR); // TODO: PSR is read to often! Read PSR only when necessary later in the ISR. PSR is reset on read! (E.g. read psr when necessary because of IR Flag, then remember that it was read, so further IR don't need to read PSR again)

	// Print Info
	if (DEBUG_PRINT) {m_can_ir_print(can_ptr->id, ir_value);}

	// init IR clear vector
	ir_reset_reg.value = 0;
	ir_reset_reg_immediately.value = 0;

	/* update Protocol Exception Statistics (because PSR Register is modified on read)	 */
	if (psr_reg.bits.PXE) { // Protocol Exception Event occurred
		can_ptr->stat.protocol.pxe++; // update statistics
	}


	/* Check and Process the individual Interrupt Flags */

	// 00: Rx FIFO 0 New Message
	if (ir_reg.bits.RF0N) {
		// Get Messages from RX_FIFO_0

		/* Target: be able to read all messages with help of the RF0N Interrupt,
		 *         it should be not necessary to poll the FIFO
		 *
		 * Action: 1. reset interrupt RF0N before reading the messages from FIFO
		 *         2. read all messages from FIFO;
		 *         Hint: if a new message arrives after reseting the IR-Flag, but before the FIFO emptied
		 *               * the corresponding message is also read from the FIFO
		 *               * the IR-Flag is set by the M_CAN
		 *               => This means, it may rarely happen for this IR routine,
		 *                  that the RF0N IR-Flag is set, but no message is in the FIFO
		 */

		// Reset the this IR Flag immediately, to not miss a new setting of the flag
		ir_reset_reg_immediately.value = 0;
		ir_reset_reg_immediately.bits.RF0N = 1;
		reg_set(can_ptr, ADR_M_CAN_IR, ir_reset_reg_immediately.value);

		// Copy RX Messages from RX FIFO_0 to a Msg Data Structure, tell FIFO_0 that Msgs are read
		can_ptr->stat.rx.msgs_ir_ok += m_can_rx_fifo_copy_msg_to_msg_list(can_ptr, 0);
	}

	// 01: Rx FIFO 0 Watermark Reached
	if (ir_reg.bits.RF0W) {
		// Reset the this IR Flag immediately, to not miss a new setting of the flag
		ir_reset_reg_immediately.value = 0;
		ir_reset_reg_immediately.bits.RF0W = 1;
		reg_set(can_ptr, ADR_M_CAN_IR, ir_reset_reg_immediately.value);

		//read all messages currently in Rx FIFO0
		can_ptr->stat.rx.msgs_ir_ok += m_can_rx_fifo_copy_msg_to_msg_list(can_ptr, 0);
	}

	// 02: Rx FIFO 0 Full
	ir_reset_reg.bits.RF0F =1;	// reset IR Flag

	// 03: Rx FIFO 0 Message Lost
    if (ir_reg.bits.RF0L) {
		if (ERROR_PRINT) {printf("[Error] M_CAN_%d: RX FIFO 0, Message Lost IR occurred\n",can_ptr->id);}
		ir_reset_reg.bits.RF0L = 1;	// reset IR Flag
	}

	// 04: Rx FIFO 1 New Message
	if (ir_reg.bits.RF1N) {
		// Get Messages from RX_FIFO_1

		/* Target: be able to read all messages with help of the RF1N Interrupt,
		 *         it should be not necessary to poll the FIFO
		 *
		 * Action: 1. reset interrupt RF1N before reading the messages from FIFO
		 *         2. read all messages from FIFO;
		 *         Hint: if a new message arrives after reseting the IR-Flag, but before the FIFO emptied
		 *               * the corresponding message is also read from the FIFO
		 *               * the IR-Flag is set by the M_CAN
		 *               => This means, it may rarely happen for this IR routine,
		 *                  that the RF1N IR-Flag is set, but no message is in the FIFO
		 */

		// Reset the this IR Flag immediately, to not miss a new setting of the flag
		ir_reset_reg_immediately.value = 0;
		ir_reset_reg_immediately.bits.RF1N = 1;
		reg_set(can_ptr, ADR_M_CAN_IR, ir_reset_reg_immediately.value);

		// Copy RX Messages from RX FIFO_1 to a Msg Data Structure, tell FIFO_1 that Msgs are read
		can_ptr->stat.rx.msgs_ir_ok += m_can_rx_fifo_copy_msg_to_msg_list(can_ptr, 1);
	}

	// 05: Rx FIFO 1 Watermark Reached
	ir_reset_reg.bits.RF1W =1;	// reset IR Flag

	// 06: Rx FIFO 1 Full
	ir_reset_reg.bits.RF1F =1;	// reset IR Flag

	// 07: Rx FIFO 1 Message Lost
	if (ir_reg.bits.RF1L) {
		if (ERROR_PRINT) {printf("[Error] M_CAN_%d: RX FIFO 1, Message Lost IR occurred\n",can_ptr->id);}
		// reset RF0N interrupt bit
		ir_reset_reg.bits.RF1L = 1;	// reset IR Flag
	}

	// 08: High Priority Message
	if (ir_reg.bits.HPM) {
		/* Hint: Each time, when a High Priority Message arrives
		*        (this means: a message ID filter element configured to generate priority event matches)
		 *       the Register HPMS is overwritten AND the IR.HPM flag is set.
		 *       In high load situation, if two high priority messages arrive back to back, it is possible that
		 *       the CPU misses to execute the interrupt before the second high priority message arrives.
		 *       As the register HPMS is overwritten, it is not possible to find out that there was an unhandled IR.HPM interrupt before
		 */

		// Reset the this IR Flag immediately, to not miss a new setting of the flag
		ir_reset_reg_immediately.value = 0;
		ir_reset_reg_immediately.bits.HPM = 1;
		reg_set(can_ptr, ADR_M_CAN_IR, ir_reset_reg_immediately.value);
		// Handle HPM Interrupt
		can_ptr->stat.rx.msgs_ir_ok += m_can_hpm_ir_handling(can_ptr);
	}

	// 09: Transmission Completed
	if (ir_reg.bits.TC) {  // Transmission Completed
		// update msg statistics
		can_ptr->stat.tx.msgs_ir_ok++;

		// update TLD statistics (Transmitter Loop Delay)
		// hint:    (1) some TLD values may be captured multiple times, and (2) some may not be captured at all!
		// reasons: (1) some frames may have the FDF==0,                and (2) some interrupts may be missed
		// hint2:   (1) to avoid duplicates in the TLD statistics, send only frames with FDF=BRS=1 and enable TDC
		// alternative: monitor TDCV periodically by polling this register (i.e. every 5 seconds)
		if ((can_ptr->bt_config.fd_ena == TRUE) && (can_ptr->bt_config.data.tdc == 1)) {

			// until the first frame is sent in FD Format + TDC, TDCV==0
			// => exclude this initial value
			if (psr_reg.bits.TDCV != 0) {
				tld_last = psr_reg.bits.TDCV - can_ptr->bt_config.data.tdc_offset;
				//printf("[M_CAN_%d] TDCV = %d, TLD = %d, TDC_offset = %d\n", can_ptr->id, psr_reg.bits.TDCV, tld_last, can_ptr->bt_config.data.tdc_offset);
				can_ptr->stat.tld.sum_cnt++;
				can_ptr->stat.tld.sum = can_ptr->stat.tld.sum + tld_last;
				if (tld_last > can_ptr->stat.tld.max) {
					can_ptr->stat.tld.max = tld_last;
				}
				if (tld_last < can_ptr->stat.tld.min) {
					can_ptr->stat.tld.min = tld_last;
				}
			}
		}
		ir_reset_reg.bits.TC = 1;	// reset IR Flag
	}

	// 10: Transmission Cancellation Finished
	ir_reset_reg.bits.TCF  =1;	// reset IR Flag

	// 11: Tx FIFO Empty
	ir_reset_reg.bits.TFE  =1;	// reset IR Flag

	// 12: Tx Event FIFO New Entry
	if(ir_reg.bits.TEFN)
	{
		// Reset the this IR Flag immediately, to not miss a new setting of the flag
		ir_reset_reg_immediately.value = 0;
		ir_reset_reg_immediately.bits.TEFN = 1;
		reg_set(can_ptr, ADR_M_CAN_IR, ir_reset_reg_immediately.value);

		// get Tx Event Fifo Elements from Message Ram and copy it to a list
		can_ptr->stat.tx_event_elements += m_can_tx_event_fifo_copy_element_to_tx_event_list(can_ptr);
	}

	// 13: Tx Event FIFO Watermark Reached
	ir_reset_reg.bits.TEFW =1;	// reset IR Flag

	// 14: Tx Event FIFO Full
	ir_reset_reg.bits.TEFF =1;	// reset IR Flag

	// 15: Tx Event FIFO Element Lost
    if (ir_reg.bits.TEFL) {
		//if (ERROR_PRINT) {printf("[Error] M_CAN_%d: TX Event FIFO Element Lost IR occurred\n",can_ptr->id);}
		ir_reset_reg.bits.TEFL =1;	// reset IR Flag
	}

	// 16: Timestamp Wraparound
    // Hint: add here code to keep track of the number of timestamp wraparounds, this is required if you evaluate the timestamps in the Elements of TX_Event_FIFO, RX Buffer, or RX FIFO
	ir_reset_reg.bits.TSW =1;	// reset IR Flag

	// 17: Message RAM Access Failure
	if (ir_reg.bits.MRAF) {
		can_ptr->stat.hw_access.mraf++; // update statistics
		//DEBUG: if (ERROR_PRINT) {printf("[Error] M_CAN_%d: Message RAM Access Failure occurred\n",can_ptr->id);}
		ir_reset_reg.bits.MRAF =1;	// reset IR Flag
	}

	// 18: Timeout Occurred
	ir_reset_reg.bits.TOO  =1;	// reset IR Flag
	//Hint: add here code to handle a timeout situation. The cause for the timeout can be known from register TOCC.TOS

	// 19: Message stored to Dedicated Rx Buffer
	if (ir_reg.bits.DRX) {
		// Reset the this IR Flag immediately, to not miss a new setting of the flag
		ir_reset_reg_immediately.value = 0;
		ir_reset_reg_immediately.bits.DRX = 1;
		reg_set(can_ptr, ADR_M_CAN_IR, ir_reset_reg_immediately.value);

		// get all Messages from dedicated rx buffers (and update statistic counter)
		can_ptr->stat.rx.msgs_ir_ok +=m_can_rx_dedicated_buffer_process_new_msg(can_ptr);
	}

	// 20: Bit Error Corrected
	if (ir_reg.bits.BEC) {
		can_ptr->stat.hw_access.bec++; // update statistics
		//if (ERROR_PRINT) {printf("[Error] M_CAN_%d: Message RAM Access - Bit Error Corrected occurred\n",can_ptr->id);}
		ir_reset_reg.bits.BEC =1;	// reset IR Flag
	}

	// 21: Bit Error Uncorrected
	if (ir_reg.bits.BEU) {
		can_ptr->stat.hw_access.beu++; // update statistics
		//if (ERROR_PRINT) {printf("[Error] M_CAN_%d: Message RAM Access - Bit Error Uncorrected occurred\n",can_ptr->id);}
		ir_reset_reg.bits.BEU =1;	// reset IR Flag
	}

	// 22: Error Logging Overflow
	ir_reset_reg.bits.ELO  =1;	// reset IR Flag

	// 23: Error Passive Status (Changed)
	if (ir_reg.bits.EP) {
		// Check if Error Passive Status is set
		if (psr_reg.bits.EP) {
			can_ptr->stat.protocol.epass++; // update statistics
		}
		can_ptr->stat.protocol.changed = TRUE;
		ir_reset_reg.bits.EP = 1;	// reset IR Flag
	}

	// 24: Error Warning Status (Changed)
	if (ir_reg.bits.EW) {
		if (psr_reg.bits.EW) { // Error Warning is set
			can_ptr->stat.protocol.ewarn++; // update statistics
		}
		can_ptr->stat.protocol.changed = TRUE;
		ir_reset_reg.bits.EW = 1;	// reset IR Flag
	}

	// 25: Bus_Off Status (Changed)
	if (ir_reg.bits.BO) {
		if (psr_reg.bits.BO) { // Bus Off is set
			can_ptr->stat.protocol.boff++; // update statistics
		}
		can_ptr->stat.protocol.changed = TRUE;
		ir_reset_reg.bits.BO   =1;	// reset IR Flag
	}

	// 26: Watchdog Interrupt
	ir_reset_reg.bits.WDI  =1;	// reset IR Flag

	// 27: Protocol Error in Arbitration Phase
	if (ir_reg.bits.PEA) {
		// set, when LEC /= 0, 7
		can_ptr->stat.protocol.lec[psr_reg.bits.LEC]++; // update statistics
		can_ptr->stat.protocol.changed = TRUE;
		ir_reset_reg.bits.PEA = 1;	// reset IR Flag
	}

	// 28: Protocol Error in Data Phase
	if (ir_reg.bits.PED) {
		// set, when DLEC /= 0, 7
		can_ptr->stat.protocol.dlec[psr_reg.bits.DLEC]++; // update statistics
		can_ptr->stat.protocol.changed = TRUE;
		ir_reset_reg.bits.PED = 1;	// reset IR Flag
	}

	// 29: Access to Reserved Address
	if (ir_reg.bits.ARA) {
		can_ptr->stat.hw_access.ara++; // update statistics
		//if (ERROR_PRINT) {printf("M_CAN_%d: Access to reserved address IR is active!\n", can_ptr->id);}
		ir_reset_reg.bits.ARA = 1;	// reset IR Flag
	}

	// Reset the IR Flags
	reg_set(can_ptr, ADR_M_CAN_IR, ir_reset_reg.value);

	// TODO: process all interrupt flags, not just reset them

} // function m_can_process_IR


/* Read new Messages from Dedicated RX Buffers */
unsigned int m_can_rx_dedicated_buffer_process_new_msg(can_struct *can_ptr)
{
	/* Description:
	 *  - reads and acknowledges new messages from dedicated Rx buffers.
	 *  - ALL messages currently in the dedicated Rx buffers are read and acknowledged
	 *
	 * Parameters:
	 *  - can_ptr: pointer to the M_CAN node, whose dedicated Rx buffer is to be checked
	 */

	int ded_buffer_index, j, msgs_read = 0;
	dword mask;
	dword new_dat_reg[2];

	// read new dat regs from M_CAN
	new_dat_reg[0] = reg_get(can_ptr, ADR_M_CAN_NDAT1);
	new_dat_reg[1] = reg_get(can_ptr, ADR_M_CAN_NDAT2);

	// scan new_dat_reg for set bits = new messages
	for (j = 0; j < 2; ++j) {

		ded_buffer_index = j*32; // init the buffer index based on the new_dat register processed currently

		mask = 1; // init mask
		while (new_dat_reg[j]) { // processes new data reg ONLY as long as there are bits set

			if (mask & new_dat_reg[j]) {

				// Check if: "dedicated Rx Buffer index > MAX_RX_BUFFER_ELEMS"
				//           If true, this means the message is stored in a memory location NOT RESERVED for ded. Rx buffers! Some other memory region was overwritten!
				if (ded_buffer_index > MAX_RX_BUFFER_ELEMS) {
					if (ERROR_PRINT) {
						printf("\n[Error] Message received in an Rx dedicated buffer %d. Memory reserved only for MAX_RX_BUFFER_ELEMS=%d", ded_buffer_index, MAX_RX_BUFFER_ELEMS);
					}
				}
				// assert did not work in an ISR => solved with if-statement above
				// assert_print_error(i <= MAX_RX_BUFFER_ELEMS, "Message received in an un-configured Rx dedicated buffer");

				m_can_rx_dedicated_buffer_copy_msg_to_msg_list(can_ptr, ded_buffer_index);
				msgs_read++;

				// Reset new_dat_bit in corresponding M_CAN
				if (j==0) { reg_set(can_ptr, ADR_M_CAN_NDAT1, mask); }
				else      { reg_set(can_ptr, ADR_M_CAN_NDAT2, mask); }
				// remove bit from local new_dat_reg[j]
				new_dat_reg[j]=new_dat_reg[j] & ~mask;

				//if (DEBUG_PRINT) printf(" Ded. RxBuffer %d: new message\n", ded_buffer_index);
			} // if
			ded_buffer_index++;        // next element
			mask = mask << 1;          // shift mask to next element
		} // while
	} // for j

	return msgs_read;
} 


/* Handle HighPriorityMessage Interrupt */
unsigned int m_can_hpm_ir_handling(can_struct *can_ptr)
{
	/* Description
	 * - The HPM interrupt signals the arrival of a high priority message
	 * - Here the interrupt is handeld exemplary in tho following way:
	 *   Copy ALL messages currently in the RX_FIFO including the High Priority message to the rx_msg_list.
	 *
	 * Parameters
	 * - can_ptr:         pointer to the M_CAN node, where the configuration should be applied
	 *
	 * Return Value
	 *   messages read from the RX FIFO
	 */

    M_CAN_HPMS_union  hpms_reg;
    unsigned int msgs_read = 0, rx_fifo_number;

    hpms_reg.value = reg_get(can_ptr, ADR_M_CAN_HPMS);

    // identify which Rx FIFO received the High Priority Message by checking HPMS.MSI
    if (hpms_reg.bits.MSI == HPMS_MSI_MESSAGE_STORED_IN_FIFO0) {
    	rx_fifo_number = 0;
    } else if (hpms_reg.bits.MSI == HPMS_MSI_MESSAGE_STORED_IN_FIFO1) {
    	rx_fifo_number = 1;
    } else {
    	if (ERROR_PRINT) {
        	if (hpms_reg.bits.MSI == HPMS_MSI_FIFO_MESSAGE_LOST) {
        		printf("\n[ERROR] [HPM Handling ISR] High Priority message lost.");
        	}
    	}

    	return msgs_read;
    }

    // copy all messages including the high priority message in the Rx FIFO to the message list
    msgs_read = m_can_rx_fifo_copy_msg_to_msg_list(can_ptr, rx_fifo_number);

    return msgs_read; // return the number messages read from RX FIFO

}



/* Handle IR of M_TTCAN Nodes */
void m_can_process_TT_IRQ(can_struct *can_ptr, dword ir_value)
{
	static M_CAN_TTIR_union ir_reg;
	static M_CAN_TTIR_union ir_reset_reg;

	// get required register values
	ir_reg.value = ir_value;

	// init IR clear vector
	ir_reset_reg.value = 0;


	/* Check and Process the individual Interrupt Flags */

	//  0 :  Start of Basic Cycle
	if (ir_reg.bits.SBC) {
		// IR is just reset
		ir_reset_reg.bits.SBC = 1;   // reset IR Flag
	}

	//  1 :  Start of Matrix Cycle
	if (ir_reg.bits.SMC) {
		// debug print: IR occured
		printf("M");
		// IR is just reset
		ir_reset_reg.bits.SMC = 1;   // reset IR Flag
	}

	//  2 :  Change of Synchronization Mode
	//  3 :  Start of Gap
	//  4 :  REG Time Mark Interrupt

	//  5 :  Trigger Time Mark Interrupt
	if (ir_reg.bits.TTMI) {
		// debug print: IR occurred
		printf("T");
		// IR is just reset
		ir_reset_reg.bits.TTMI = 1;   // reset IR Flag
	}

	//  6 :  Stop Watch Event
	//  7 :  Global Time Wrap
	//  8 :  Global Time Discontinuity
	//  9 :  Global Time Error
	//  10:  Tx Count Underflow
	//  11:  Tx Count Overflow
	//  12:  Scheduling Error 1
	//  13:  Scheduling Error 2
	//  14:  Error Level Changed
	//  15:  Initialization Watch Trigger
	//  16:  Watch Trigger
	//  17:  Application Watchdog
	//  18:  Configuration Error

	// Reset the ALL IR Flags, as long as not all IRs are handled individually
	ir_reset_reg.value = 0xFFFFFFFF;
	reg_set(can_ptr, ADR_M_CAN_TTIR, ir_reset_reg.value);

} // function m_can_process_IR

